
/*  Various programs concerning the El-Fish .fsh file format.
    Copyright (C) 2004 Vidar Holen

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


package holen.vidar.elfish;

import java.io.*;
import java.util.*;
import java.awt.*;
import java.awt.image.*;
import java.awt.event.*;
import javax.swing.*;

public class Fish { 
    String name;
    ArrayList[] seqs;
    Image icon, detailed;
    int dirs=0;
    
    public Fish(String name) {
        this.name=name;
        seqs=new ArrayList[6*7];
        for(int i=0; i<seqs.length; i++) 
            seqs[i]=new ArrayList();
    }

    public ArrayList getSequence(int d, int n) {
        return seqs[seqToPos(d,n)];
    }
    public ArrayList getRaw(int r) {
        return seqs[r];
    }

    public void addFrame(FishFrame f) { 
        ArrayList s=getSequence(f.dir, f.seq);
        int i=0;
        while(i<s.size() && ((FishFrame)s.get(i)).no<f.no) i++;
        s.add(i,f);

        if(f.dir>dirs) dirs=f.dir;

        System.out.println("Added "+i+": "+f);
    }
 
    protected int seqToPos(int d, int n) {
        return d*7+n;
    }

    public int getDirections() {
        return dirs;
    }
    public int invertDirection(int d) {
        int n=(dirs/2)+d;
        return n;
    }
    public void finish() {
        dirs++;
        System.out.println("Fish "+name+" finalized, with "+dirs+" directions.");
    }

    
    public static Fish loadFish(String file) throws Exception {
        RandomAccessFile raf=new RandomAccessFile(file,"r");
        StringBuffer sb=new StringBuffer();
        String name;
        int w,h;

        raf.seek(0x0A);
        for(int i=0; i<8; i++) {
            int c=raf.read();
            if(c==0) break;
            else sb.append((char)c);
        }
        name=sb.toString().toLowerCase();
        Fish fish=new Fish(name);

        raf.seek(0x8B);
        int det=iflip(raf.readInt());
        int icon=iflip(raf.readInt());
        int frames=iflip(raf.readInt());

        /*
        raf.seek(det);
        raf.skipBytes(26);
        int w=sflip(raf.readShort());
        int h=sflip(raf.readShort());
        decodeFrame(w,h,raf,name+"-detailed.ppm");
        
        raf.seek(icon);
        raf.skipBytes(26);
        w=sflip(raf.readShort());
        h=sflip(raf.readShort());
        decodeFrame(w,h,raf,name+"-icon.ppm");
        */
        
        int n=0;
        int[] fd=new int[8];
        while(frames!=0) {
            raf.seek(frames);
            for(int i=0; i<5; i++)
                fd[i]=iflip(raf.readInt());
            for(int i=5; i<8; i++)
                fd[i]=sflip(raf.readShort());
            w=sflip(raf.readShort());
            h=sflip(raf.readShort());

            System.out.println("#"+n+" Len="+h(fd[0])+"\t p="+h(fd[2])+"\t n="+h(fd[3])+"\t "+w+","+h);
            Image m=decodeFrame(w,h,raf);
            FishFrame ff=new FishFrame(w,h,fd[1],fd[5],(short)fd[6],(short)fd[7],m);
            fish.addFrame(ff);
            
            frames=fd[3];
            n++;
        }
        raf.close();
        fish.finish();
        return fish;
    }
    public static String h(int n) {
        return Integer.toHexString(n);
    }
    public static void p(String s, int n) {
        System.out.println(s+" "+h(n)+" ("+n+")");
    }
    public static int iflip(int n) {
        return flip(n,4);
    }
    public static int sflip(int n) {
        return flip(n,2);
    }
    public static int flip(int n, int b) {
        int a=0;
        for(int i=0; i<b; i++) {
            a=(a<<8) | (n&0xFF);
            n=n>>8;
        }
        return a;
    }
    public static String pad(int n, int p) {
        String s=Integer.toString(n);
        while(s.length()<p)
            s="0"+s;
        return s;
    }
    public static BufferedImage decodeFrame(int w, int h, 
            RandomAccessFile raf) throws Exception {
        int[] lbuf=new int[w];
        Arrays.fill(lbuf,0x00FFFFFF);
        BufferedImage bi=new BufferedImage(w,h,BufferedImage.TYPE_INT_ARGB);
        
        int nh=0;
        while(nh<h-1) {
            int len=sflip(raf.readShort());
            int pos=sflip(raf.readShort());
            int col;

            if(pos<0x8000) {
                bi.setRGB(0,nh,w,1,lbuf,0,w);
                Arrays.fill(lbuf,0x00FFFFFF);
                nh++;
            } else {
                pos=0x10000-pos;
            }
            col=pos;

            for(int i=0; i<len; i++) 
                lbuf[pos++]=palette[raf.read()] | (0xff000000);
        }
        bi.setRGB(0,nh,w,1,lbuf,0,w);

        return bi;
    }

    final public static int[] palette=new int[] { 
        0x000000, 0x7c8090, 0x30384c, 0xa8a4ac, 0x000000, 0xfc5454, 0xe8fc60,
        0xa80000, 0x587088, 0xf0f0f0, 0x707080, 0x7898a0, 0x606060, 0x000000,
        0xd00000, 0xfcfcfc, 0x000000, 0x282828, 0x505050, 0x7c7c7c, 0xa4a4a4,
        0xd0d0d0, 0xf8f8f8, 0x280000, 0x282800, 0x002800, 0x002828, 0x000028,
        0x280028, 0x502828, 0x505028, 0x285028, 0x285050, 0x282850, 0x502850,
        0x7c5050, 0x7c7c50, 0x507c50, 0x507c7c, 0x50507c, 0x7c507c, 0xa47c7c,
        0xa4a47c, 0x7ca47c, 0x7ca4a4, 0x7c7ca4, 0xa47ca4, 0xd0a4a4, 0xd0d0a4,
        0xa4d0a4, 0xa4d0d0, 0xa4a4d0, 0xd0a4d0, 0xf8d0d0, 0xf8f8d0, 0xd0f8d0,
        0xd0f8f8, 0xd0d0f8, 0xf8d0f8, 0x500000, 0x502800, 0x505000, 0x285000,
        0x005000, 0x005028, 0x005050, 0x002850, 0x000050, 0x280050, 0x500050,
        0x500028, 0x7c2828, 0x7c5028, 0x7c7c28, 0x507c28, 0x287c28, 0x287c50,
        0x287c7c, 0x28507c, 0x28287c, 0x50287c, 0x7c287c, 0x7c2850, 0xa45050,
        0xa47c50, 0xa4a450, 0x7ca450, 0x50a450, 0x50a47c, 0x50a4a4, 0x507ca4,
        0x5050a4, 0x7c50a4, 0xa450a4, 0xa4507c, 0xd07c7c, 0xd0a47c, 0xd0d07c,
        0xa4d07c, 0x7cd07c, 0x7cd0a4, 0x7cd0d0, 0x7ca4d0, 0x7c7cd0, 0xa47cd0,
        0xd07cd0, 0xd07ca4, 0xf8a4a4, 0xf8d0a4, 0xf8f8a4, 0xd0f8a4, 0xa4f8a4,
        0xa4f8d0, 0xa4f8f8, 0xa4d0f8, 0xa4a4f8, 0xd0a4f8, 0xf8a4f8, 0xf8a4d0,
        0x7c0000, 0x7c3000, 0x7c6400, 0x647c00, 0x307c00, 0x007c00, 0x007c30,
        0x007c64, 0x00647c, 0x00307c, 0x00007c, 0x30007c, 0x64007c, 0x7c0064,
        0x7c0030, 0xa42828, 0xa45c28, 0xa48c28, 0x8ca428, 0x5ca428, 0x28a428,
        0x28a45c, 0x28a48c, 0x288ca4, 0x285ca4, 0x2828a4, 0x5c28a4, 0x8c28a4,
        0xa4288c, 0xa4285c, 0xd05454, 0xd08454, 0xd0b854, 0xb8d054, 0x84d054,
        0x54d054, 0x54d084, 0x54d0b8, 0x54b8d0, 0x5484d0, 0x5454d0, 0x8454d0,
        0xb854d0, 0xd054b8, 0xd05484, 0xf87c7c, 0xf8b07c, 0xf8e07c, 0xe0f87c,
        0xb0f87c, 0x7cf87c, 0x7cf8b0, 0x7cf8e0, 0x7ce0f8, 0x7cb0f8, 0x7c7cf8,
        0xb07cf8, 0xe07cf8, 0xf87ce0, 0xf87cb0, 0xa40000, 0xa44000, 0xa48400,
        0x84a400, 0x40a400, 0x00a400, 0x00a440, 0x00a484, 0x0084a4, 0x0040a4,
        0x0000a4, 0x4000a4, 0x8400a4, 0xa40084, 0xa40040, 0xd02828, 0xd06c28,
        0xd0b028, 0xb0d028, 0x6cd028, 0x28d028, 0x28d06c, 0x28d0b0, 0x28b0d0,
        0x286cd0, 0x2828d0, 0x6c28d0, 0xb028d0, 0xd028b0, 0xd0286c, 0xf85454,
        0xf89454, 0xf8d854, 0xd8f854, 0x94f854, 0x54f854, 0x54f894, 0x54f8d8,
        0x54d8f8, 0x5494f8, 0x5454f8, 0x9454f8, 0xd854f8, 0xf854d8, 0xf85494,
        0xd00000, 0xd04c00, 0xd09c00, 0xb4d000, 0x68d000, 0x18d000, 0x00d034,
        0x00d080, 0x00d0d0, 0x0080d0, 0x0034d0, 0x1800d0, 0x6800d0, 0xb400d0,
        0xd0009c, 0xd0004c, 0xf82828, 0xf87828, 0xf8c428, 0xe0f828, 0x90f828,
        0x44f828, 0x28f85c, 0x28f8ac, 0x28f8f8, 0x28acf8, 0x285cf8, 0x4428f8,
        0x9028f8, 0xe028f8, 0xf828c4, 0xf82878 
    };

}


